unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, OpenGLContext, Forms, Controls, Graphics,
  Dialogs, ExtCtrls, StdCtrls, gl, glu;

type
  ToglVector3 = record
    x, y, z: GLfloat;
  end;

  TPaletteItem = record
    Value: GLFloat;
    Color: TColor;
  end;

  ToglColor4 = record
    R, G, B, A: GLfloat;
  end;

  TPalette = array of TPaletteItem;

  TCalculateEvent = function(x, y: Double): Double;

  { TForm1 }

  TForm1 = class(TForm)
    btnCopyToClipboard: TButton;
    cbWireFrame: TCheckBox;
    cbAxes: TCheckBox;
    cbBoundingBox: TCheckBox;
    cbDataPoints: TCheckBox;
    cmbFunction: TComboBox;
    cbColorMaterial: TCheckBox;
    cbLight: TCheckBox;
    cbuseColorPalette: TCheckBox;
    cbLightPosFixed: TCheckBox;
    cmbDataPointSymbols: TComboBox;
    Label1: TLabel;
    lblDatapointSymbols: TLabel;
    lblFunction: TLabel;
    OpenGLControl: TOpenGLControl;
    Panel1: TPanel;
    procedure btnCopyToClipboardClick(Sender: TObject);
    procedure cbAxesChange(Sender: TObject);
    procedure cbBoundingBoxChange(Sender: TObject);
    procedure cbColorMaterialChange(Sender: TObject);
    procedure cbDataPointsChange(Sender: TObject);
    procedure cbTransparentChange(Sender: TObject);
    procedure cmbDataPointSymbolsChange(Sender: TObject);
    procedure cmbFunctionChange(Sender: TObject);
    procedure cbLightChange(Sender: TObject);
    procedure cbLightPosFixedChange(Sender: TObject);
    procedure cbUseColorPaletteChange(Sender: TObject);
    procedure cbWireFrameChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure OpenGLControlMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OpenGLControlMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure OpenGLControlMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure OpenGLControlPaint(Sender: TObject);
    procedure OpenGLControlResize(Sender: TObject);
  private
    { private declarations }
    FRotAngleX, FRotAngleY: Double;
    FShiftX, FShiftY: Double;
    FDistance: Double;
    FMouseDownPt: TPoint;
    FPrevMousePt: TPoint;
    FGLInitDone: Boolean;
    FSceneInitDone: boolean;
    FPalette: TPalette;
    FXMin, FXMax, FYMin, FYMax, FZMin, FZMax: GLFloat;
    FNx, FNy: Integer;
    FOnCalculate: TCalculateEvent;
    FSphere: PGLUQuadric;
    FCone: PGLUQuadric;

    FVertices: array of ToglVector3;
    FNormals: array of ToglVector3;
    FColors: array of ToglColor4;
    FIndices: array of Integer;

    procedure ApplyProjection;
    procedure CalcColors;
    procedure CalcFunc;
    procedure DrawAxes;
    procedure DrawBoundingBox;
    procedure DrawDataPoints;
    procedure DrawFunc;

    procedure InitFunc;
    procedure InitGL;
    procedure InitLights;
    procedure InitPalette;

    procedure ScreenToWorld(xs, ys, zs: GLFloat; out xw, yw, zw: GLFloat);
    function ScreenToWorldX(xs: GLfloat): GLfloat;
    function ScreenToWorldY(ys: GLfloat): GLfloat;
    function ScreenToWorldZ(zs: GLfloat): GLfloat;

    procedure WorldToScreen(xw, yw, zw: GLFloat; out xs, ys, zs: GLFloat);
    function WorldToScreenX(xw: GLfloat): GLfloat;
    function WorldToScreenY(yw: GLfloat): GLfloat;
    function WorldToScreenZ(zw: GLfloat): GLfloat;
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  Math,
  Clipbrd, IntfGraphics, GraphType;

{ Some demo functions }

function PolynomialFunc(x, y: Double): Double;
begin
  Result := 0.2*(2*x*x*x - x*x - 2*(x+1)*y - 0.5*y*y);
end;


function SincFunc(x, y: Double): Double;
var
  r: Double;
begin
  if (x = 0) and (y = 0) then
    Result := 1
  else begin
    r := sqrt(sqr(x) + sqr(y));
    Result := sin(r) / r;
  end;
end;


function EllipticParaboloid(x, y: Double): Double;
begin
  Result := sqr(x) + sqr(y);
end;


function HyperbolicParaboloid(x, y: Double): Double;
begin
  Result := sqr(x) - sqr(y);
end;


{ Colors }

procedure ClearPalette(var P: TPalette);
begin
  SetLength(P, 0);
end;

// Values must be added in ascending order, this is not checked.
procedure AddToPalette(var P: TPalette; AValue: GLFloat; AColor: TColor);
begin
  SetLength(P, Length(P) + 1);
  P[High(P)].Value := AValue;
  P[High(P)].Color := AColor;
end;

{ borrowed from TAChart }
procedure InterpolateRGB(AColor1, AColor2: TColor; ACoeff: Double;
  out R, G, B: GLfloat);
begin
  if ACoeff < 0 then ACoeff := 0 else if ACoeff > 1.0 then ACoeff := 1.0;
  R := EnsureRange(Red(AColor1) * (1 - ACoeff) + Red(AColor2) * ACoeff, 0, 255) / 255;
  G := EnsureRange(Green(AColor1) * (1 - ACoeff) + Green(AColor2) * ACoeff, 0, 255) / 255;
  B := EnsureRange(Blue(AColor1) * (1 - ACoeff) + Blue(AColor2) * ACoeff, 0, 255) / 255;
end;

procedure InterpolateColor(const P: TPalette; AValue: GLFloat;
  out R, G, B: GLfloat);
var
  i: Integer;
  c: Double;
begin
  if AValue < P[0].Value then begin
    R := Red(P[0].Color) / 255;
    G := Green(P[0].COlor) / 255;
    B := Blue(P[0].Color) / 255;
  end else
  if AValue > P[High(P)].Value then begin
    R := Red(P[High(P)].Color) / 255;
    G := Green(P[High(P)].Color) / 255;
    B := Blue(P[High(P)].Color) / 255;
  end else
    for i:= 0 to High(P) - 1 do
      if (AValue >= P[i].Value) and (AValue <= P[i+1].Value) then begin
        c := (AValue - P[i].Value) / (P[i+1].Value - P[i].Value);
        InterpolateRGB(P[i].Color, P[i+1].Color, c, R, G, B);
        exit;
      end;
end;


{ Vector math }

procedure Normalize(var A: ToglVector3);
const
  EPS = 1E-5;
var
  len: Double;
begin
  len := sqrt(A.X*A.X + A.Y*A.Y + A.Z*A.Z);
  if abs(len) < EPS then exit;
  A.x := A.x / len;
  A.y := A.y / len;
  A.z := A.z / len;
end;

function Cross(A, B: ToglVector3): ToglVector3;
begin
  Result.x := A.y * B.z - A.z * B.y;
  Result.y := A.z * B.x - A.x * B.z;
  Result.z := A.x * B.y - A.y * B.z;
end;

operator +(A, B: ToglVector3): TOglVector3;
begin
  Result.x := A.x + B.x;
  Result.y := A.y + B.y;
  Result.z := A.z + B.z;
end;

operator -(A, B: ToglVector3): TOglVector3;
begin
  Result.x := A.x - B.x;
  Result.y := A.y - B.y;
  Result.z := A.z - B.z;
end;

function Normal(A, B, C: ToglVector3): ToglVector3;
begin
  Result := Cross(B-A, C-A);
  Normalize(Result);
end;




{ OpenGL export }

procedure OpenGLToBitmap(ABitmap: TBitmap);
const
  GL_BGRA = $80E1;
var
  intfImg: TLazIntfImage;
  viewport: array[0..3] of GLInt;
  rawImg: TRawImage;
  bmpHandle, maskHandle: THandle;
begin
  if ABitmap = nil then
    raise Exception.Create('[OpenGLToBitmap] Bitmap must not be nil.');

  // Query size of the viewport
  glGetIntegerv(GL_VIEWPORT, @viewport);

  // Prepare a raw image
  rawImg.Init;
  rawImg.Description.Init_BPP32_B8G8R8A8_M1_BIO_TTB(viewport[2], viewport[3]);
  rawImg.Description.LineOrder := riloBottomToTop;
  rawImg.CreateData(false);

  // Query image data from OpenGL
  glReadPixels(0, 0, viewport[2], viewport[3], GL_BGRA, GL_UNSIGNED_BYTE, rawImg.Data);

  // Create LazIntfImage from raw image
  intfImg := TLazIntfImage.Create(viewport[2], viewport[3]);
  try
    intfImg.SetRawImage(rawImg);

    // Convert LazIntfImage to Bitmap
    intfImg.CreateBitmaps(bmpHandle, maskHandle);
    ABitmap.Handle := bmpHandle;
    ABitmap.MaskHandle := maskHandle;
  finally
    intfImg.Free;
  end;
end;


{ TForm1 }

procedure TForm1.ApplyProjection;
const
  VIEWANGLE = 40;
  NEAR_CLIPDIST = 0.1;
  FAR_CLIPDIST = 100;
var
  w, h: Integer;
  d: Double;
begin
  w := OpenGLControl.Width;
  h := OpenGLControl.Height;
  if h = 0 then h := 1;

  // Set viewport to window dimensions
  glViewport(0, 0, w, h);

  // Reset coordinate system
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;

  // Create perspective projection
  gluPerspective(
    VIEWANGLE,      // field of view (degrees) in y direction
    w/h,            // aspect ratio
    NEAR_CLIPDIST,  // distance to near clipping plane
    FAR_CLIPDIST    // distance to far clipping plane
  );

  // Reset current matrix to MODELVIEW
  glMatrixMode(GL_MODELVIEW);
end;


procedure TForm1.btnCopyToClipboardClick(Sender: TObject);
var
  bmp: TBitmap;
begin
  bmp := TBitmap.Create;
  try
    OpenGLToBitmap(bmp);
    Clipboard.Assign(bmp);
  finally
    bmp.Free;
  end;
end;


procedure TForm1.cbAxesChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbBoundingBoxChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbColorMaterialChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;


procedure TForm1.cbDataPointsChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;

procedure TForm1.cmbFunctionChange(Sender: TObject);
begin
  case cmbFunction.ItemIndex of
    0: begin
         FOnCalculate := @PolynomialFunc;
         FXMin := -2;
         FXMax := +2;
         FYMin := -2;
         FYMax := +2;
       end;
    1: begin
         FOnCalculate := @SincFunc;
         FXMin := -20;
         FXMax := +20;
         FYMin := -20;
         FYMax := +20;
       end;
    2: begin
         FOnCalculate := @EllipticParaboloid;
         FXMin := -3;
         FXMax := +3;
         FYMin := -3;
         FYMax := +3;
       end;
    3: begin
         FOnCalculate := @HyperbolicParaboloid;
         FXMin := -5;
         FXMax := +5;
         FYMin := -5;
         FYMax := +5;
       end;
  end;
  FSceneInitDone := false;
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbLightChange(Sender: TObject);
begin
  CalcFunc;
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbLightPosFixedChange(Sender: TObject);
begin
  FSceneInitDone := false;
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbTransparentChange(Sender: TObject);
begin
  CalcColors;
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbuseColorPaletteChange(Sender: TObject);
begin
  CalcColors;
  OpenGLControl.Invalidate;
end;


procedure TForm1.cbWireFrameChange(Sender: TObject);
begin
  FSceneInitDone := false;  // Need a new spacing
  OpenGLControl.Invalidate;
end;

procedure TForm1.cmbDataPointSymbolsChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;


procedure TForm1.DrawAxes;
const
  AXIS_LENGTH = 100;
var
  x0, y0, z0: GLFloat;
begin
  // Calculate position of origin in scaled coordinate system
  WorldToScreen(0, 0, 0, x0, y0, z0);

  glLineWidth(3);
  glEnable(GL_BLEND);  // for line smoothing

  glColor3f(1, 0, 0);   // x axis --> red
  glBegin(GL_LINES);
    glVertex3f(x0, y0, z0);
    glVertex3f(x0 + AXIS_LENGTH, y0, z0);
  glEnd;

  glColor3f(0, 0, 1);  // OpenGL y axis (= Model z axis) --> blue
  glBegin(GL_LINES);
    glVertex3f(x0, y0, z0);
    glVertex3f(x0, y0 + AXIS_LENGTH, z0);
  glEnd;

  glColor3f(0, 1, 0);  // OpenGL z axis (= Model y axis) green
  glBegin(GL_LINES);
    glVertex3f(x0, y0, z0);
    glVertex3f(x0, y0, z0 + AXIS_LENGTH);
  glEnd;

  glDisable(GL_BLEND);
end;


procedure TForm1.DrawBoundingBox;
var
  x1, x2, y1, y2, z1, z2: GLFloat;
begin
  WorldToScreen(FXMin, FYMin, FZMin, x1, y1, z1);
  WorldToScreen(FXMax, FYMax, FZMax, x2, y2, z2);

  glLineWidth(1);
  glEnable(GL_BLEND);  // for line smoothing

  glColor4f(0, 0, 0, 0.2);
  glBegin(GL_LINE_STRIP);
    glVertex3f(x1, y1, z1);
    glVertex3f(x2, y1, z1);
    glVertex3f(x2, y1, z2);
    glVertex3f(x1, y1, z2);
    glVertex3f(x1, y1, z1);
  glEnd;
  glBegin(GL_LINE_STRIP);
    glVertex3f(x1, y2, z1);
    glVertex3f(x2, y2, z1);
    glVertex3f(x2, y2, z2);
    glVertex3f(x1, y2, z2);
    glVertex3f(x1, y2, z1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x1, y1, z1);
    glVertex3f(x1, y2, z1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x1, y1, z2);
    glVertex3f(x1, y2, z2);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x2, y1, z1);
    glVertex3f(x2, y2, z1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x2, y1, z2);
    glVertex3f(x2, y2, z2);
  glEnd;

  glDisable(GL_BLEND);
end;


procedure TForm1.DrawDataPoints;
const
  SIZE = 0.02;
  N = 30;

  procedure DrawNormalizedCube;
  begin
    glBegin(GL_POLYGON);
      glNormal3f(0.0, 0.0, 1.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(-1.0, -1.0, 1.0);
      glVertex3f(1.0, -1.0, 1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glNormal3f(0.0, 0.0, -1.0);
      glVertex3f(1.0, 1.0, -1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(-1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(-1.0, 1.0, -1.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(-1.0, -1.0, 1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glNormal3f(1.0, 0.0, 0.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(1.0, -1.0, 1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glNormal3f(0.0, 1.0, 0.0);
      glVertex3f(-1.0, 1.0, -1.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glNormal3f(0.0, -1.0, 0.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(1.0, -1.0, 1.0);
      glVertex3f(-1.0, -1.0, 1.0);
    glEnd;
  end;

  procedure DrawDataPoint(x, y, z: GLFloat);
  var
    xs, ys, zs: GLFloat;
  begin
    WorldToScreen(x, y, z, xs, ys, zs);
    glPushMatrix;
      glTranslatef(xs, ys, zs);
      case cmbDataPointSymbols.ItemIndex of
        0: gluSphere(FSphere, SIZE, 18, 18);
        1: gluCylinder(FCone, SIZE, 0, SIZE*1.4, 18, 10);
        2: begin
             glScalef(SIZE, SIZE, SIZE);
             DrawNormalizedCube;
           end;
      end;
    glPopMatrix;
  end;

var
  i: Integer;
  x, y, z: GLFloat;
begin
  if not Assigned(FOnCalculate) then
    exit;

  RandSeed := 0;
  glColor3f(0, 0, 1);  // data points in blue
  for i := 0 to N - 1 do begin
    x := Random * (FXMax - FXMin) + FXMin;
    y := Random * (FYMax - FYMin) + FYMin;
    z := FOnCalculate(x, y) + (Random * 2 - 1) * (FZMax - FZMin)*0.05;   // 5% scatter
    DrawDataPoint(x, y, z);
  end;
end;


procedure TForm1.CalcColors;
var
  i: Integer;
  z: Double;
begin
  if not cbUseColorPalette.Checked then begin
    SetLength(FColors, 0);
    exit;
  end;

  SetLength(FColors, Length(FVertices));
  for i := 0 to High(FVertices) do begin
    z := FVertices[i].z;
    // z is scaled to run from -0.5 to +0.5. The colors in the palette, however
    // have assigned values between 0 and 1 --> add 0.5 to z.
    InterpolateColor(FPalette, z + 0.5,
      FColors[i].R, FColors[i].G, FColors[i].B);
  end;
end;


procedure TForm1.CalcFunc;

  procedure CalcDerivatives(x, y: GLfloat; out dzdx, dzdy: GLfloat);
  const
    dxs: GLfloat = 1E-2;
    dys: GLfloat = 1E-2;
  var
    x1, x2, y1, y2, z1, z2: Double;
    zs1, zs2: Double;
  begin
    x1 := ScreenToWorldX(WorldToScreenX(x) - dxs);
    z1 := FOnCalculate(x1, y);
    zs1 := WorldToScreenZ(z1);

    x2 := ScreenToWorldX(WorldToScreenX(x) + dxs);
    z2 := FOnCalculate(x2, y);
    zs2 := WorldToScreenZ(z2);

    dzdx := (zs2 - zs1) / (2 * dxs);

    y1 := ScreenToWorldY(WorldToScreenY(y) - dys);
    z1 := FOnCalculate(x, y1);
    zs1 := WorldToScreenZ(z1);

    y2 := ScreenToWorldY(WorldToScreenY(y) + dys);
    z2 := FOnCalculate(x, y2);
    zs2 := WorldToScreenZ(z2);

    dzdy := (zs2 - zs1) / (2 * dys);
  end;

var
  dx, dy: GLFloat;
  i, j: Integer;
  x, y, z: Double;
  xs, ys, zs: GLfloat;
  dzdx, dzdy: GLfloat;
  idx: Integer;
  c: TColor;
  len: Double;
  A, B, N: ToglVector3;
begin
  if not Assigned(FOnCalculate) then
    exit;

  dx := (FXMax - FXMin) / (FNx - 1);
  dy := (FYMax - FYMin) / (FNy - 1);

  SetLength(FVertices, FNx * FNy);
  idx := 0;
  for j := 0 to FNy-1 do begin
    y := FYMin + j * dy;
    for i := 0 to FNx-1 do begin
      // Vertex
      x := FXMin + i * dx;
      z := FOnCalculate(x, y);
      WorldToScreen(x, y, z, xs, ys, zs);
      FVertices[idx].x := xs;
      FVertices[idx].y := ys;
      FVertices[idx].z := zs;
  {
      // Normals at vertices
      CalcDerivatives(x, y, dzdx, dzdy);  // normal vectors are (-dz/dx, -dz/dy, 1)
      len := sqrt(sqr(dzdx) + sqr(dzdy) + 1);  // normalize to 1
      FNormals[idx].x := -dzdx / len;
      FNormals[idx].y := -dzdy / len;
      FNormals[idx].z := 1.0 / len;
   }
      inc(idx);
    end;
  end;

  // Calculate normals
  SetLength(FNormals, FNx * FNy);
  for i:=0 to High(FVertices) do
    FNormals[i] := Normal(FVertices[i], FVertices[i+1], FVertices[i+FNx]);

  { Connect vertices in clockwise orientation
             i, j+1    i+1, j+1
                +---->---+
                |        |
                ^        v
                |        |
                +----<---+
               i,j    i+1, j
  }
  SetLength(FIndices, (FNx-1) * (FNy-1) * 4);
  idx := 0;
  for j := 0 to FNy-2 do
    for i := 0 to FNx-2 do begin
      FIndices[idx  ] := j*FNx + i;
      FIndices[idx+1] := (j+1) * FNx + i;
      FIndices[idx+2] := (j+1) * FNx + i + 1;
      FIndices[idx+3] := j*FNx + i + 1;
      inc(idx, 4);
    end;

  CalcColors;
end;

procedure TForm1.DrawFunc;
begin
  if not Assigned(FOnCalculate) then
    exit;

  if cbWireFrame.Checked then begin
    glLineWidth(1);
    glPolygonMode(GL_FRONT, GL_LINE);
    glPolygonMode(GL_BACK, GL_LINE);
  end else begin
    glPolygonMode(GL_FRONT, GL_FILL);
    glPolygonMode(GL_BACK, GL_FILL);
  end;

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  if cbUseColorPalette.Checked then
    glEnableClientState(GL_COLOR_ARRAY)
  else
    glColor3f(0.7, 0.3, 0.2);

  glVertexPointer(3, GL_FLOAT, 0, @FVertices[0]);
  glNormalPointer(GL_FLOAT, 0, @FNormals[0]);
  if cbUseColorPalette.Checked then
    glColorPointer(4, GL_FLOAT, 0, @FColors[0]);

  glDrawElements(GL_QUADS, Length(FIndices), GL_UNSIGNED_INT, @FIndices[0]);

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  if cbUseColorPalette.Checked then
    glDisableClientState(GL_COLOR_ARRAY);
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  cmbFunctionChange(nil);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  FDistance := 4;
  FRotAngleX := 20;
  FRotAngleY := 30;

  FSphere := gluNewQuadric();
  gluQuadricDrawStyle(FSphere, GLU_FILL);

  FCone := gluNewQuadric();
  gluQuadricDrawStyle(FCone, GLU_FILL);
end;


procedure TForm1.InitFunc;
var
  i, j: Integer;
  x, y, z: GLFloat;
begin
  if not Assigned(FOnCalculate) then
    exit;

  if cbWireFrame.Checked then begin
    FNx := 41;
    FNy := 41;
  end else begin
    FNx := 201;
    FNy := 201;
  end;

  ////////////////////
//  FNx := 21;
//  FNy := 21;

  FZMin := 1E30;
  FZMax := -1E30;
  for j := 0 to FNy - 1 do begin
    y := FYMin + (FYMax - FYMin) * j / (FNy - 1);
    for i := 0 to FNx - 1 do begin
      x := FXMin + (FXMax - FXMin) * i / (FNx - 1);
      z := FOnCalculate(x, y);
      FZMin := Min(FZMin, z);
      FZMax := Max(FZMax, z);
    end;
  end;
end;


procedure TForm1.InitGL;
begin
  if FGlInitDone then
    exit;

  FGLInitDone := true;

  glClearColor(0.5, 0.5 ,0.5, 1.0);   // sets background color
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);             // the type of depth test to do
  glEnable(GL_DEPTH_TEST);            // enables depth testing
  glShadeModel(GL_SMOOTH);            // enables smooth color shading

  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
end;


procedure TForm1.InitLights;
const
  AMBIENT: array[0..3] of GLfloat  = (0.3, 0.3, 0.3, 1.0);    // ambient light
  DIFFUSE: array[0..3] of GLfloat  = (0.7, 0.7, 0.7, 1.0);    // diffuse light
  SPECULAR: array[0..3] of GLfloat = (1.0, 1.0, 1.0, 1.0);    // specular light
  LIGHT_POS: array[0..3] of GLfloat = (0.0, 1000.0, 1000.0, 0.0);   // directional light
begin
  // set up light colors (ambient, diffuse, specular)
  glLightfv(GL_LIGHT0, GL_AMBIENT, @AMBIENT);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, @DIFFUSE);
  glLightfv(GL_LIGHT0, GL_SPECULAR, @SPECULAR);

  // position the light
  glLightfv(GL_LIGHT0, GL_POSITION, @LIGHT_POS);

  // MUST enable each light source after configuration
  glEnable(GL_LIGHT0);
end;


procedure TForm1.InitPalette;
begin
  ClearPalette(FPalette);
  AddToPalette(FPalette, 0.00, clBlack);
  AddToPalette(FPalette, 0.10, clBlue);
  AddToPalette(FPalette, 0.40, clRed);
  AddToPalette(FPalette, 0.70, clGreen);
  AddToPalette(FPalette, 0.80, clYellow);
  AddToPalette(FPalette, 1.00, clWhite);
end;


procedure TForm1.OpenGLControlMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMouseDownPt := Point(X, Y);
  FPrevMousePt := FMouseDownPt;
end;


procedure TForm1.OpenGLControlMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
const
  SPEED = 0.5;
  DELTA = 0.003;
var
  f: Double;
begin
  f := 1.0;
  if ssShift in Shift then
    f := f * 5.0;
  if ssLeft in Shift then begin
    if FMouseDownPt.Y > OpenGLControl.Height div 2 then begin
      FRotAngleY := FRotAngleY + (X - FPrevMousePt.X)*SPEED*f;
      FRotAngleX := FRotAngleX + (Y - FPrevMousePt.Y)*SPEED*f;
    end
    else begin
      FRotAngleY := FRotAngleY - (X - FPrevMousePt.X)*SPEED*f;
      FRotAngleX := FRotAngleX - (Y - FPrevMousePt.Y)*SPEED*f
    end;
    FPrevMousePt := Point(X, Y);
    OpenGLControl.Invalidate;
  end else
  if ssRight in Shift then begin
    FShiftX := FShiftX + (x - FPrevMousePt.X)*DELTA*f;
    FShiftY := FShiftY + (FPrevMousePt.Y - y)*DELTA*f;
    FPrevMousePt := Point(X, Y);
    OpenGLControl.Invalidate;
  end;
end;


procedure TForm1.OpenGLControlMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
const
  SPEED = 0.0005;
var
  f: Double;
begin
  if (ssCtrl in Shift) then
    f := -1.0
  else
    f := 1.0;
  if (ssShift in Shift) then
    f := f * 2;
  FDistance := FDistance * (1.0 + f * WheelDelta * SPEED);
  OpenGLControl.Invalidate;
end;


procedure TForm1.OpenGLControlPaint(Sender: TObject);
begin
  if not OpenGLControl.MakeCurrent then
    exit;

  if not FSceneInitDone then begin
    InitPalette;
    InitFunc;
    InitGL;
    if cbLightPosFixed.Checked then InitLights;
    ApplyProjection;
    CalcFunc;
    FSceneInitDone := true;
  end;

  glMatrixMode(GL_MODELVIEW);

  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;             { clear the modelview matrix }

  { viewing transformation }
  glTranslatef(FDistance*FShiftX, FDistance*FShiftY, -FDistance);
  glRotatef(FRotAngleX, 1, 0, 0);
  glRotatef(FRotAngleY, 0, 1, 0);

  // Rotate scene so that z points upward
  glRotatef(-90, 1, 0, 0);

  if not cbLightPosFixed.Checked then InitLights;
  if cbLight.Checked then glEnable(GL_LIGHTING) else glDisable(GL_LIGHTING);

  if cbColorMaterial.Checked then
    glEnable(GL_COLOR_MATERIAL)
  else
    glDisable(GL_COLOR_MATERIAL);

  // Function
  DrawFunc;

  // Data points
  if cbDataPoints.Checked then DrawDataPoints;

  // Bounding box
  if cbBoundingBox.Checked then DrawBoundingBox;

  // Axes
  if cbAxes.Checked then DrawAxes;

  OpenGLControl.SwapBuffers;
end;


procedure TForm1.OpenGLControlResize(Sender: TObject);
begin
  if FSceneInitDone and OpenGLControl.MakeCurrent then
    ApplyProjection;
end;


function TForm1.ScreenToWorldX(xs: GLfloat): GLfloat;
begin
  Result := (xs + 1) * (FXMax - FXMin) + FXMin;
end;

function TForm1.ScreenToWorldY(ys: GLfloat): GLfloat;
begin
  Result :=(ys + 1) * (FYMax - FYMin) + FYMin;
end;

function TForm1.ScreenToWorldZ(zs: GLfloat): GLfloat;
begin
  Result :=(zs + 0.5) * (FZMax - FZMin) + FZMin;
end;

procedure TForm1.ScreenToWorld(xs, ys, zs: GLFloat; out xw, yw, zw: GLFloat);
begin
  xw := (xs + 1) * (FXMax - FXMin) + FXMin;
  yw := (ys + 1) * (FYMax - FYMin) + FYMin;
  zw := (zs + 0.5) * (FZMax - FZMin) + FZMin;
end;


procedure TForm1.WorldToScreen(xw, yw, zw: GLFloat; out xs, ys, zs: GLFloat);
begin
  xs := (xw - FXMin) / (FXMax - FXMin) * 2 - 1;    // -1 ... +1
  ys := (yw - FYMin) / (FYMax - FYMin) * 2 - 1;    // -1 ... +1
  zs := (zw - FZMin) / (FZMax - FZMin) - 0.5;      // -0.5 ... 0.5
end;

function TForm1.WorldToScreenX(xw: GLfloat): GLfloat;
begin
  Result := (xw - FXMin) / (FXMax - FXMin) * 2 - 1;    // -1 ... +1
end;

function TForm1.WorldToScreenY(yw: GLfloat): GLfloat;
begin
  Result := (yw - FYMin) / (FYMax - FYMin) * 2 - 1;    // -1 ... +1
end;

function TForm1.WorldToScreenZ(zw: GLfloat): GLfloat;
begin
  Result := (zw - FZMin) / (FZMax - FZMin) - 0.5;    // -1 ... +1
end;

end.

